from nautobot.apps.jobs import Job
from nautobot.apps.jobs import ObjectVar, BooleanVar
from nautobot.dcim.models import Device

from nautobot_obe.utils.device import delete_device

import json

name = "orange de-provisioning jobs"


class DeleteDevice(Job):
    """Deletes a device.

    We want to keep track of historical devices so we will not really delete devices from nautobot but
    rather remove any service references and set the status to "Decommissioning".

    As an option it will be possible to completely remove the device from the database.
    """

    device = ObjectVar(
        model=Device,
        label="Device",
        description="Choose the device that should be removed.",
        required=True,
        query_params={
            "depth": 0,
        },
    )

    force_remove = BooleanVar(
        label="Force delete from DB",
        description="If enabled then the device will be deleted from Nautobot instead of decommissioned.",
        default=False,
    )

    keep_mgmt_ip = BooleanVar(
        label="Keep mgmt IP",
        description="If enabled then the management IP will not be released and will not become available for new services.",
        default=False,
    )

    template_name = "orange/job_base.html"

    class Meta:
        """extra class info."""

        name = "Delete CPE"
        description = """This will decommission or delete a device.
        
        Decommissioning a device means:
        
        - the management IP address of a device will be released (deleted) so that it becomes available again in the ip pool
        - any service ids will be removed from the asset_tag field and from the Tags field
        - the status will be changed to "Decommissioned"
        
        if the option "Keep mgmt IP" is selected then the management IP address will be unlinked from the device but it will not be released.
        
        By default a device will never be deleted from Nautobot.
        However, if "force delete" is selected then the Device will be completely deleted from Nautobot. 
        """
        field_order = [
            "device",
            "force_remove",
            "keep_mgmt_ip",
        ]

    def run(self, device, force_remove, keep_mgmt_ip):
        """Runs the job."""
        try:
            output = delete_device(device, force_remove, keep_mgmt_ip)
            return json.dumps(
                {
                    "data": output,
                    "failed": False,
                }
            )
        except Exception as e:
            return json.dumps({"failed": True, "error": str(e)})
