from nautobot.apps.jobs import Job
from nautobot.apps.jobs import StringVar, ChoiceVar, ObjectVar, BooleanVar
from nautobot.tenancy.models import Tenant
from nautobot.ipam.models import VRF, VLAN
from nautobot.extras.models import Tag
from nautobot.dcim.models import Device
from nautobot.dcim.models import DeviceType
from nautobot.dcim.models import Location

from nautobot_obe.utils.vrf import (
    create_new_vrf,
    VRF_TYPES,
)
from nautobot_obe.utils.ipaddress import (
    request_new_ips,
    INTERNET_LAN_PREFIX_LENGTH,
)

from nautobot_obe.utils.device import create_new_device

from nautobot_obe.utils.vlan import (
    voipt_search_first_free_vlan,
    reserve_voipt_vlan,
    security_package_search_first_free_vlan,
    reserve_security_package_vlan,
)

import json

name = "orange provisioning jobs"


class CreateVrf(Job):
    """Creates a new VRF and matching route-targets.

    Input:
      - Customer (account id)
      - VRF name
      - VRT type (Any, Hub, Spoke)

    Output:
      {
          "account": {
              "id": <accountid>,
              "name": <account name>,
          }
          "vrf": <vrf name>,
          "type": <vrf type>,
          "rd": <vrf rd>,
          "rt_import": [ <rt> ],
          "rt_export": [ <rt> ],
      }
    """

    tenant = ObjectVar(
        model=Tenant,
        label="Account ID",
        display_field="description",
    )

    template_name = "orange/job_create_vrf.html"

    # template_name = "orange/job_create_vrf.html"
    vrf = StringVar(
        label="VRF",
        description="VRF name. If you specify the index then the VRF should not yet exist. If you omit the index, then the next available index will be automatically generated.",
        regex="^[A-Z0-9_]+[^_]$",
    )

    vrf_type = ChoiceVar(
        label="VRF Type",
        choices=VRF_TYPES,
        description="This selection defines the route-targets",
    )

    # TODO: this should be required to have form validation however if it's hidden
    #       then the form post will not work
    parent_vrf = ObjectVar(
        model=VRF,
        label="HUB VRF",
        null_option="-- select the HUB vrf --",
        required=False,
        query_params={"cf_cf_vrf_type": "HUB"},
    )

    class Meta:
        """Job description."""

        name = "Create VRF"
        description = """This job creates a new VRF for a customer.
        
        The account (tenant) already has to exist, otherwise the job will fail.
        
        A VRF name may contain an index (= end with _0001) or it can be added
        without index. If there is an index then the VRF should not yet exist.
        If there is no index then an index will be created automatically and
        it will be the first available index for that VRF.
        
        An "ANY" vrf will have 1 new route-target that is set for both import
        and export.
        
        A "HUB" vrf will have 1 new route-target for both import + expart. And
        also a new RT for import only and a new RT for export only.
        
        A "SPOKE vrf will only inherit the route-targets from the parent, we
        don't create a route-target for both import + export.
        You have to select a valid HUB VRF when creating a new spoke vrf.
        """
        field_order = ["tenant", "vrf", "vrf_type", "parent_vrf"]

    def run(
        self,
        tenant: Tenant,
        vrf: str,
        vrf_type: str,
        parent_vrf: VRF | None,
    ):
        """Creates a new VRF and route-targets."""
        try:
            new_vrf: VRF = create_new_vrf(
                name=vrf,
                vrf_type=vrf_type,
                account=tenant,
                parent_vrf=parent_vrf,
            )
            return json.dumps(
                {
                    "data": dict(
                        vrf=new_vrf.name,
                        type=vrf_type,
                        rd=new_vrf.rd,
                        rt_import=[str(rt.name) for rt in new_vrf.import_targets.all()],
                        rt_export=[str(rt.name) for rt in new_vrf.export_targets.all()],
                        account=dict(accountid=tenant.name, account_name=tenant.description),
                    ),
                    "created": True,
                }
            )
        except Exception as e:
            return json.dumps({"created": False, "error": str(e)})


class CreateCpeDevice(Job):
    """Creates a new device and automatically assigns a new loopback IP.

    It should be enough to provide the VT and device type. The NIS API will be queried
    to find addtional details.
    The Tenant short name should be filled in for the hostname to be generated automatically
    """

    template_name = "orange/job_create_device.html"

    serviceid = ObjectVar(
        model=Tag,
        label="Service ID",
        description="VT reference of the first service on a CPE, this VT will also be linked to the loopback IP.",
        required=True,
    )

    devicetype = ObjectVar(
        model=DeviceType,
        label="Device Type",
        description="The device type is required so that the correct device template can be applied.",
        required=True,
    )

    customer_shortname = StringVar(
        label="Customer short name",
        description="Customer short name.",
        regex="^[a-z0-9]+$",
        required=True,
    )

    countrycode = StringVar(
        label="Country code",
        description="Single letter country code (B=Belgium, etc).",
        regex="^[a-z]$",
        required=True,
        default="b",
    )

    zipcode = StringVar(
        label="Zipcode",
        description="Location zip code for this CPE, the first 3 digits will be used in the hostname.",
        regex="^[0-9]{3,4}$",
        required=True,
    )

    generated_hostname = StringVar(
        label="Hostname",
        description="Proposed hostname (if the index is not present then it will be added automatically.)",
        regex="^[a-z0-9\-]+$",
        required=True,
    )

    tenant = StringVar(
        label="tenant",
        required=False,
    )

    class Meta:
        """extra class info."""

        name = "Create CPE Device"
        description = """This will create a new CPE device and generates a new loopback IP.
        
        The purpose is to provide only the Service ID and the Device Type. Nautobot will find the customer shortname from its own database and the country and zipcode from the the NIS API.
        
        The hostname will automatically be generated if the customer short name + country code + zipcode are found.
        It is possible at this moment to manually override the generated fields.
        
        Hostname format = [customer short name]-[1 letter country][3 digit zipcode]-[index]
        """
        field_order = [
            "serviceid",
            "devicetype",
            "customer_shortname",
            "countrycode",
            "zipcode",
            "generated_hostname",
            "tenant",
        ]

    def run(
        self,
        serviceid,
        devicetype,
        customer_shortname,
        countrycode,
        zipcode,
        generated_hostname,
        tenant,
        # location,
        # devicename,
    ):
        """Runs the job."""

        try:
            output = create_new_device(serviceid, tenant, devicetype, generated_hostname)
            return json.dumps(
                {
                    "data": output,
                    "created": True,
                }
            )
        except Exception as e:
            return json.dumps({"created": False, "error": str(e)})


class CreateIPAddressForService(Job):
    """Creates IP addresses for CI or IPVPN services.

    inputs:
      - Service ID
      - Customer (account id)

    outputs:
      {

      }
    """

    serviceid = ObjectVar(
        model=Tag,
        label="Service ID",
        description="VT reference",
    )

    # tenant = ObjectVar(
    #     model=Tenant,
    #     label="Account ID",
    #     display_field="description",
    # )
    tenant = StringVar(
        label="tenant",
        required=False,
    )

    device = ObjectVar(
        model=Device,
        label="CPE router",
        query_params={
            "depth": 0,
        },
    )

    lan_prefix_length = ChoiceVar(
        label="Internet LAN Prefix length",
        required="False",
        choices=INTERNET_LAN_PREFIX_LENGTH,
        description="Internet LAN range prefix length (for internet services), leave default for IP-VPN. The service type will be detected based on the Service ID description field.",
        default="29",
    )

    template_name = "orange/job_fill_account_id.html"

    class Meta:
        """Job description."""

        name = "Create IP addresses for service"
        description = """This job assigns IP addresses for a service.
        
        The account (tenant) already has to exist, otherwise the job will fail.
        
        The service type should be known via the Tag.
        In case of IPVPN:
            - search free loopback IP
            - search free IPVPN P2P
        In case of CI:
            - search free loopback IP
            - search free internet P2P
            - search free internet LAN /29
            
        For Loopback IP => a CPE should have only 1 single management loopback IP, if it already has one
        then re-use it, otherwise assign it.
        """
        field_order = ["serviceid", "tenant"]

    def run(
        self,
        serviceid: Tag,
        # tenant: Tenant,
        tenant: str,
        device: Device,
        lan_prefix_length: str,
    ):
        """Creates new ip addresses."""
        try:
            ip_addresses_output = request_new_ips(serviceid, tenant, device, lan_prefix_length)
            return json.dumps(
                {
                    "data": ip_addresses_output,
                    "created": True,
                }
            )
        except Exception as e:
            return json.dumps({"created": False, "error": str(e)})


class CreateVOIPTrunk(Job):
    """Assigns resources for VOIP trunk.

    inputs:
      - Service ID
      - Customer (account id)

    This will lookup the next free pre-defined VOIP Trunk vlan,
    Assign the VT to the Vlan + prefixes.
    And return the Vlan + VT + prefixes + ip addresses
    """

    serviceid = ObjectVar(
        model=Tag,
        label="Service ID",
        description="Voip Trunking VT reference",
    )

    # tenant = ObjectVar(
    #     model=Tenant,
    #     label="Account ID",
    #     display_field="description",
    # )

    tenant = StringVar(
        label="tenant",
        required=False,
    )

    vlan = ObjectVar(
        model=VLAN,
        label="Select a vlan",
        null_option="-- leave blank for automatic vlan selection --",
        required=False,
        query_params={
            "vlan_group": "VOIP Trunking",
            "tenant__isnull": "true",
            "tags__isnull": "true",
            "status": "Reserved",
        },
    )

    template_name = "orange/job_fill_account_id.html"

    class Meta:
        """Job description."""

        name = "Create VOIP Trunk vlan and ip addresses"
        description = """This job assigns a vlan and matching ip addresses to a VOIP Trunking VT.
        
        All vlans and ip addresses are pre-defined according to the engineering excel file.        
        The first available Vlan with status "reserved" will be used, we keep the same order
        as the excel file ("2500:2599", "2700:2999", "2400:2499", "2600:2699")
        
        You can override automatic vlan selection if needed, only "reserved" vlans can be
        selected.
        """
        field_order = ["serviceid", "tenant", "device", "vlan"]

    def run(
        self,
        serviceid: Tag,
        # tenant: Tenant,
        tenant: str,
        vlan: VLAN,
    ):
        """Creates a new VRF and route-targets."""
        try:
            selected_vlan = vlan if vlan else voipt_search_first_free_vlan()
            result = reserve_voipt_vlan(selected_vlan, serviceid, tenant)
            return json.dumps(
                {
                    "data": result,
                    "created": True,
                }
            )
        except Exception as e:
            return json.dumps({"created": False, "error": str(e)})


class CreateSecurityPackage(Job):
    """Assigns resources for Security Package.

    inputs:
      - Service ID
      - Customer (account id)

    This will lookup the next free pre-defined Security Package vlan,
    Assign the VT to the Vlan + prefixes.
    And return the Vlan + VT + prefixes + ip addresses
    """

    serviceid = ObjectVar(
        model=Tag,
        label="Service ID",
        description="Security Package VT reference",
    )

    # tenant = ObjectVar(
    #     model=Tenant,
    #     label="Account ID",
    #     display_field="description",
    # )

    tenant = StringVar(
        label="tenant",
        required=False,
    )

    vlan = ObjectVar(
        model=VLAN,
        label="Select a vlan",
        null_option="-- leave blank for automatic vlan selection --",
        required=False,
        query_params={
            "vlan_group": "Security Package",
            "tenant__isnull": "true",
            "tags__isnull": "true",
            "status": "Reserved",
        },
    )

    lan_prefix_length = ChoiceVar(
        label="Internet LAN Prefix length",
        required="False",
        choices=INTERNET_LAN_PREFIX_LENGTH,
        description="Internet LAN range prefix length, from the reserved Security Package LAN range.",
        default="29",
    )

    template_name = "orange/job_fill_account_id.html"

    class Meta:
        """Job description."""

        name = "Create Security Package vlan and ip addresses"
        description = """This job assigns a vlan and matching ip addresses to a Security Package VT.
        
        All vlans and ip addresses are pre-defined according to the engineering excel file.        
        The first available Vlan with status "reserved" will be used, we keep the same order
        as the excel file ("2300:2400")
        
        You can override automatic vlan selection if needed, only "reserved" vlans can be
        selected.
        """
        field_order = ["serviceid", "tenant", "vlan"]

    def run(
        self,
        serviceid: Tag,
        # tenant: Tenant,
        tenant: str,
        vlan: VLAN,
        lan_prefix_length: str,
    ):
        """Job for creating a new Security Package."""
        try:
            selected_vlan = vlan if vlan else security_package_search_first_free_vlan()
            result = reserve_security_package_vlan(selected_vlan, serviceid, tenant, lan_prefix_length)
            return json.dumps(
                {
                    "data": result,
                    "created": True,
                }
            )
        except Exception as e:
            return json.dumps({"created": False, "error": str(e)})
