import re
from nautobot.ipam.models import IPAddress
from nautobot.ipam.models import RouteTarget
from nautobot.tenancy.models import Tenant
from nautobot.dcim.models import Device
from nautobot.dcim.models import DeviceType
from nautobot.extras.models import Tag
from django.db.models import Q
from nautobot.extras.models import Status
from nautobot.extras.models import Role
from nautobot.dcim.models import Location
from nautobot.ipam.models import Interface
from datetime import datetime, timedelta
from django.utils import timezone

from nautobot_obe.utils.ipaddress import create_new_management_ip

import logging
import re

logger = logging.getLogger(__name__)

UNKNOWN_LOCATION = "-unknown-"


def update_status_to_offline(last_seen_days: int = 60) -> dict:
    """Updates the status of devices to Offline if last_seen date exceeds the threshold.

    We need to find devices that:
        - are active and have a field cf_last_seen set and cf_last_seen is older than the cutoff_date
        - are active without a cf_last_seen value defined but have a create date older than the cutoff_date
    """
    cutoff_date = datetime.now().date() - timedelta(days=last_seen_days)
    offline_status = Status.objects.filter(name="Offline").first()
    active_status = Status.objects.filter(name="Active").first()

    unseen_devices = (
        Device.objects.exclude(_custom_field_data__cf_last_seen__isnull=True)
        .exclude(_custom_field_data__cf_last_seen=None)
        .filter(status=active_status)
        .filter(_custom_field_data__cf_last_seen__lte=cutoff_date)
    )
    devices_without_last_seen_field = Device.objects.filter(
        Q(_custom_field_data__cf_last_seen__isnull=True) | Q(_custom_field_data__cf_last_seen=None)
    ).filter(status=active_status)

    updated_devices = []
    failed_devices = []

    for device in list(unseen_devices) + list(devices_without_last_seen_field):
        try:
            device.status = offline_status
            device.save()
            updated_devices.append(device)
        except:
            failed_devices.append(device)
            continue

    return {
        "cutoff_date": str(cutoff_date),
        "updated_devices": updated_devices,
        "failed_devices": failed_devices,
    }


def device_exists(devicename: str, serviceid: Tag) -> list[Device]:
    """Checks if a device already exists based on name or tag."""
    devices = Device.objects.filter(Q(name__iexact=devicename) | Q(tags=serviceid) | Q(asset_tag=serviceid)).all()
    return devices


def generate_device_name(base_name: str) -> str:
    """Generates or completes the base_devicename.

    - if base_name already ends with an index -> do nothing
    - if the base does not have an index then search the DB for similar
      hostnames and add an index
    """
    if re.match(".*\-\d+$", base_name):
        return base_name
    idx = 1
    devices = Device.objects.filter(name__icontains=base_name).all()
    for device in devices:
        try:
            _idx = int(device.name.split("-")[-1])
            if _idx >= idx:
                idx = _idx + 1
        except Exception:
            pass
    return f"{base_name}-{str(idx).zfill(2)}"


def get_lowest_loopback_interface(device: Device):
    """Finds and returns the lowest loopback interface."""

    def mysorter(intf):
        """Used to sort on the loopback number id."""
        m = re.match(r".*\D(\d+)$", intf.name)
        return m.groups()[0]

    loopbacks = device.interfaces.filter(type="virtual").filter(name__icontains="loopback").all()
    if not loopbacks:
        return None
    loopbacks = sorted(loopbacks, key=mysorter)
    return loopbacks[0]


def assign_device_primary_ip(device: Device, ipaddress: IPAddress) -> Interface:
    """Sets the primary IP address for a device.

    The IP will first be added to the primary loopback interface.
    If it doesn't exist then this function will fail.
    """
    lp = get_lowest_loopback_interface(device)
    if not lp:
        raise Exception("the device does not have any loopback interfaces")
    lp.ip_addresses.add(ipaddress)
    device.primary_ip4 = ipaddress
    device.save()

    return lp


def create_new_device(
    serviceid: Tag,
    tenant: str,
    # tenant: Tenant,
    # location: Location,
    devicetype: DeviceType,
    devicename: str,
) -> Device:
    """Creates a new device and also creates and assigns the loopback IP.

    The device will be tagged with the VT reference of the primary IP,
    this is normally the very first service assigned added to the device.

    Args:
    ----
        serviceid (Tag): tag that will be assigned for loopback
        tenant (Tenant): will be assigned to device
        location (Location): required but we always use a default for now
        devicetype (DeviceType): needed to create a new device
        devicename (str): new name of the device

    Returns:
    -------
        output (dict)
    """
    new_device_name = generate_device_name(devicename)
    existing_devices = device_exists(new_device_name, serviceid)
    if existing_devices:
        raise Exception(
            f"Device(s) with name '{new_device_name}' or tag '{serviceid.name}' already exists: {existing_devices}"
        )

    PlannedStatus = Status.objects.filter(name="Planned").first()
    ReservedStatus = Status.objects.filter(name="Reserved").first()
    TenantObject = Tenant.objects.filter(name=tenant).first()
    LocationObject = Location.objects.filter(name=UNKNOWN_LOCATION).first()
    CPERole = Role.objects.filter(name="CPE").first()

    new_device = Device.objects.create(
        name=new_device_name,
        device_type=devicetype,
        status=PlannedStatus,
        role=CPERole,
        location=LocationObject,
        tenant=TenantObject,
        asset_tag=serviceid.name,
    )
    new_device.tags.add(serviceid)

    new_lp_ip = create_new_management_ip(
        status=ReservedStatus,
        tenant=TenantObject,
        serviceid=serviceid,
        device=new_device,
    )
    lp_intf = assign_device_primary_ip(new_device, new_lp_ip)

    return dict(
        service_id=serviceid.name,
        name=new_device.name,
        mgmt_ip=str(new_lp_ip.address),
        mgmt_intf=lp_intf.name,
        device_type=devicetype.display,
    )


def delete_device(device: Device, force_remove: bool, keep_mgmt_ip: bool):
    """Deletes/decommissions a device."""
    # delete the management ip if keep_mgmt_ip=False
    # delete the device if force_remove=True
    # remove asset tag + tags + update description + status
    output = dict(
        force_removal=force_remove, keep_mgmt_ip=keep_mgmt_ip, removed=False, updated=False, mgmt_ip_removed=False
    )
    DecommissioningStatus = Status.objects.filter(name="Decommissioning").first()
    DecommissionedRole = Role.objects.filter(name="DECOMMISSIONED").first()
    if keep_mgmt_ip is False:
        for intf in [i for i in device.interfaces.all() if i.mgmt_only is True]:
            for ip_link in intf.ip_address_assignments.all():
                ip_link.ip_address.delete()
        output["mgmt_ip_removed"] = True
    if force_remove is True:
        device.delete()
        output["removed"] = True
    if force_remove is False:
        # unlink any IP's from interfaces
        for intf in device.interfaces.all():
            for ip_link in intf.ip_address_assignments.all():
                ip_link.delete()
        device.status = DecommissioningStatus
        device.asset_tag = None
        device.tags.clear()
        device.tenant = None
        device.role = DecommissionedRole
        device.save()
        output["updated"] = True
    return output
