from nautobot.tenancy.models import Tenant
from nautobot.ipam.models import Prefix
from nautobot.ipam.models import IPAddress
from nautobot.extras.models import Tag
from nautobot.dcim.models import Device
from nautobot.dcim.models import Interface
from nautobot.extras.models import Role
from django.db.models import Q
from nautobot.extras.models import Status


import netaddr
import logging
import re

logger = logging.getLogger(__name__)

INTERNET_LAN_PREFIX_LENGTH = (
    ("28", "/28"),
    ("29", "/29"),
    ("30", "/30"),
    ("31", "/31"),
    ("32", "/32"),
)

ROLE_IPVPN_LOOPBACK = "IPVPN_CE_LOOPBACK"
ROLE_IPVPN_P2P = "IPVPN_P2P"
ROLE_CI_P2P = "INTERNET_P2P"
ROLE_CI_LAN = "INTERNET_LAN"


def get_lowest_loopback_interface(device: Device):
    """Finds and returns the lowest loopback interface."""

    def mysorter(intf):
        """Used to sort on the loopback number id."""
        m = re.match(r".*\D(\d+)$", intf.name)
        return m.groups()[0]

    loopbacks = device.interfaces.filter(type="virtual").filter(name__icontains="loopback").all()
    if not loopbacks:
        return None
    loopbacks = sorted(loopbacks, key=mysorter)
    return loopbacks[0]


def assign_device_primary_ip(device: Device, ipaddress: IPAddress) -> Interface:
    """Sets the primary IP address for a device.

    The IP will first be added to the primary loopback interface.
    If it doesn't exist then this function will fail.
    """
    lp = get_lowest_loopback_interface(device)
    if not lp:
        raise Exception("the device does not have any loopback interfaces")
    lp.ip_addresses.add(ipaddress)
    device.primary_ip4 = ipaddress
    device.save()

    return lp


def get_prefix_with_role(role: str) -> Prefix:
    """Finds the prefixes with lowest utilization that matches the given role.

    Only a single prefix is returned.

    Args:
    ----
        role (str) : name of the role, ex. INTERNET_P2P
    """
    role_in_db = Role.objects.filter(name=role).first()
    if not role_in_db:
        raise Exception(f"role '{role}' was not found")

    prefixes_with_role = Prefix.objects.filter(role=role_in_db).filter(Q(type="pool") | Q(type="container")).all()
    if not prefixes_with_role:
        raise Exception(f"no prefixes found with role '{role}'")

    prefixes_with_role = sorted(prefixes_with_role, key=lambda pfx: pfx.get_utilization())
    logger.debug(f"prefixes with role '{role_in_db.name} found: {prefixes_with_role}")

    return prefixes_with_role[0]


def get_free_ip_from_prefix(prefix: Prefix) -> IPAddress:
    """Finds the first free IP address in a prefix.

    The prefix is supposed to be a 'pool' type.
    """
    free_ips = prefix.get_available_ips()
    if not free_ips:
        raise Exception(f"unable to find a free IP address in prefix '{prefix}'")
    free_ip = next(free_ips.__iter__())
    return free_ip


def get_free_subnet_from_prefix(prefix: Prefix, prefix_length: int = 30) -> netaddr.IPNetwork:
    """Finds a free subnet with disired prefix_length from a prefix."""
    free_subnets: netaddr.IPSet = prefix.get_available_prefixes()
    if not free_subnets:
        raise Exception(f"unable to find a free subnet in prefix '{prefix}'")

    for cidr in free_subnets.iter_cidrs():
        if cidr.prefixlen <= prefix_length:
            return netaddr.IPNetwork(f"{cidr.network}/{prefix_length}")

    raise Exception(f"unable to find a subnet with the desired prefix length '{prefix_length}'")


def get_device_primary_ip(device: Device) -> IPAddress | None:
    """Finds the primary IPv4 address of a devices.

    First the primary_ip field is checked.
    If nothing found then an IP address of any loopback interface is checked.
    """
    if device.primary_ip:
        logger.debug("device already has a primary ip address")
        return device.primary_ip

    lp = get_lowest_loopback_interface(device)

    lp_ip_addresses = lp.ip_addresses.all()
    logger.debug(f"device loopback addresses found: {lp_ip_addresses}")
    if lp_ip_addresses:
        logger.debug("loopback interface found with ip address, assigning it to the device")
        device.primary_ip4 = lp_ip_addresses[0]
        device.save()
        return device.primary_ip

    return None


def loopback_ipaddress_exists_with_tag(serviceid: Tag) -> bool:
    """If an ip address already exists with this tag then return True."""
    addresses = (
        IPAddress.objects.filter(mask_length=32)
        .filter(Q(tags=serviceid) | Q(description__contains=serviceid))
        .all()
        .distinct()
    )
    if addresses:
        return True
    return False


def prefix_exists_with_tag(serviceid: Tag, prefix_length=30) -> bool:
    """If a prefix already exists with this service id (either in tags or in description) then return True."""
    prefixes = (
        Prefix.objects.filter(prefix_length=prefix_length)
        .filter(Q(tags=serviceid) | Q(description__contains=serviceid))
        .all()
        .distinct()
    )
    if prefixes:
        return True
    return False


def create_new_management_ip(
    status: Status,
    tenant: Tenant,
    serviceid: Tag,
    device: Device,
) -> IPAddress:
    """Creates a new loopback IP for management purposes."""
    if loopback_ipaddress_exists_with_tag(serviceid) is True:
        raise Exception(f"A loopback IP address with this serviceid {serviceid.name} already exists")

    lp_pfx = get_prefix_with_role(ROLE_IPVPN_LOOPBACK)
    logger.debug(f"LOOPBACK prefixes found: lp={lp_pfx}")
    new_ip = get_free_ip_from_prefix(lp_pfx)
    logger.debug(f"Free loopback IP found: {new_ip}/32")

    mgmt_ip = IPAddress.objects.create(
        address=f"{new_ip}/32",
        type="host",
        status=status,
        tenant=tenant,
        description=f"{serviceid.name} - {device.name}",
    )
    mgmt_ip.tags.add(serviceid)
    return mgmt_ip


def request_new_ips(
    serviceid: Tag,
    # tenant: Tenant,
    tenant_name: str,
    device: Device,
    lan_prefix_length: str,
):
    """Finds and assigns IP addresses for a service."""
    mgmt_ip: IPAddress = get_device_primary_ip(device)
    logger.debug(f"device management ip address = {mgmt_ip}")

    tenant = Tenant.objects.filter(name=tenant_name).first()

    service_type = "internet" if "internet" in serviceid.description.lower() else "ipvpn"
    logger.debug(f"selecting service type '{service_type}' based on the tag description")
    ReservedStatus = Status.objects.filter(name="Reserved").first()

    output = {
        "service_id": serviceid.name,
        "service_type": service_type,
        "device": device.name,
        "loopback_created": False,
    }

    # we always search for a free mgmt Loopback IP, no matter what the service is
    # this IP belongs to the CPE
    # If the CPE already has a management IP assigned, the we will use this and
    # there is no need to create a new one
    if not mgmt_ip:
        mgmt_ip = create_new_management_ip(
            status=ReservedStatus,
            tenant=tenant,
            serviceid=serviceid,
            device=device,
        )
        output["loopback_created"] = True

        # try to assign the loopback IP to the device
        assign_device_primary_ip(device, mgmt_ip)

    output["loopback_ip"] = str(mgmt_ip.address)

    if prefix_exists_with_tag(serviceid=serviceid, prefix_length=30) is True:
        raise Exception(f"A P2P ip with serviceid {serviceid.name} already exists")

    # TODO: we really should check this in the correct parent prefixes because both
    # P2P as LAN could be /30 for example
    if lan_prefix_length != 30 and prefix_exists_with_tag(serviceid=serviceid, prefix_length=lan_prefix_length) is True:
        raise Exception(f"A LAN range with serviceid {serviceid.name} already exists")

    if service_type == "internet":
        p2p = get_prefix_with_role(ROLE_CI_P2P)
        lan = get_prefix_with_role(ROLE_CI_LAN)
        logger.debug(f"CI prefixes found: p2p={p2p}, lan={lan}")
        new_p2p_subnet = get_free_subnet_from_prefix(p2p, 30)

        logger.debug(f"Free P2P subnet found:{new_p2p_subnet}")
        logger.debug(f"required lan prefix length: {lan_prefix_length}")
        new_lan_subnet = get_free_subnet_from_prefix(lan, int(lan_prefix_length))
        logger.debug(f"Free LAN subnet found:{new_lan_subnet}")
        output["p2p"] = str(new_p2p_subnet)
        output["lan"] = str(new_lan_subnet)

    if service_type == "ipvpn":
        p2p = get_prefix_with_role(ROLE_IPVPN_P2P)
        logger.debug(f"IPVPN prefixes found: p2p={p2p}")
        new_p2p_subnet = get_free_subnet_from_prefix(p2p, 30)
        logger.debug(f"Free P2P subnet found:{new_p2p_subnet}")
        new_lan_subnet = None
        output["p2p"] = str(new_p2p_subnet)

    for _new_prefix in [new_p2p_subnet, new_lan_subnet]:
        if _new_prefix:
            new_subnet = Prefix.objects.create(
                prefix=str(_new_prefix),
                type="network",
                status=ReservedStatus,
                description=serviceid.name,
                tenant=tenant,
            )
            new_subnet.tags.add(serviceid)

    return output
