"""Functions related to VRFs."""

import re

from nautobot.ipam.models import VLAN, VLANGroup, Prefix
from nautobot.extras.models import Status, Tag
from nautobot.tenancy.models import Tenant
from nautobot_obe.utils.ipaddress import get_prefix_with_role, get_free_subnet_from_prefix

import logging

# from django.db.models import Value, IntegerField
# from django.db.models.functions import Replace, Cast

logger = logging.getLogger(__name__)

# these are the ranges that are used for provisioning according to Engineering xls file
# we fill the ranges in the order defined in the xls, although in theory we can use
# any Reserved vlan
VOIP_TRUNK_VLAN_RANGES = ["2500:2599", "2700:2999", "2400:2499", "2600:2699"]
SECURITY_PACKAGE_VLAN_RANGES = ["1300:1400"]

VLAN_GROUP_VOIP_TRUNKING = "VOIP Trunking"
VLAN_GROUP_SECURITY_PACKAGE = "Security Package"
SECURITY_PACKAGE_LAN_PREFIX_ROLE = "SECURITY_PACKAGE_INTERNET"


def get_all_reserved_voip_trunking_vlans() -> list[VLAN]:
    """Returns all reserved VOIP Trunking vlans."""
    vlan_group_obj = VLANGroup.objects.filter(name=VLAN_GROUP_VOIP_TRUNKING).get()
    ReservedStatus = Status.objects.filter(name="Reserved").first()

    all_vlans = VLAN.objects.filter(
        vlan_group=vlan_group_obj,
        status=ReservedStatus,
        tenant__isnull=True,
        tags__isnull=True,
    ).all()
    if not all_vlans:
        raise Exception(
            "No free voip trunking vlans found. Status should be 'reserved', tenant and tags should be empty."
        )
    logger.debug(f"All free vlans found: {all_vlans}")

    return all_vlans


def get_all_reserved_security_package_vlans() -> list[VLAN]:
    """Returns all reserved VOIP Trunking vlans."""
    vlan_group_obj = VLANGroup.objects.filter(name=VLAN_GROUP_SECURITY_PACKAGE).get()
    ReservedStatus = Status.objects.filter(name="Reserved").first()

    all_vlans = VLAN.objects.filter(
        vlan_group=vlan_group_obj,
        status=ReservedStatus,
        tenant__isnull=True,
        tags__isnull=True,
    ).all()
    if not all_vlans:
        raise Exception(
            "No free security package vlans found. Status should be 'reserved', tenant and tags should be empty."
        )
    logger.debug(f"All free vlans found: {all_vlans}")

    return all_vlans


def voipt_search_first_free_vlan() -> VLAN:
    """Finds the first abailable VOIPT vlan.

    We loop over the VOIP_TRUNK_VLAN_RANGES until we can find the first free (reserved) vlan.
    This Vlan is returned.
    """
    all_free_vlans = get_all_reserved_voip_trunking_vlans()
    available_vlans = [int(vlan.vid) for vlan in all_free_vlans]
    vlan_found = None
    for _range in VOIP_TRUNK_VLAN_RANGES:
        _from, _to = _range.split(":")
        for _vid in range(int(_from), int(_to) + 1):
            if _vid in available_vlans:
                vlan_found = _vid
                break

    if not vlan_found:
        raise Exception(f"Unable to find a free (reserved) vlan within these ranges: {VOIP_TRUNK_VLAN_RANGES}")

    vlan_obj = next(filter(lambda x: x.vid == vlan_found, list(all_free_vlans)))

    return vlan_obj


def reserve_voipt_vlan(
    vlan: VLAN,
    serviceid: Tag,
    # tenant: Tenant,
    tenant_name: str,
) -> dict:
    """Block a VOIPT vlan and matching prefixes and ip addresses.

    All vlans, prefixes and ip-addresses are pre-defined in Nautobot for VOIP
    Trunking. We just need to allocate the Tag to each resource in order to link
    it with the service.

    Actions to do:

    - vlan:
        * status update to Active
        * description update (format: "VTxxxx | VOIP Trunking")
        * add tenant
        * add Tag

    - prefix:
        * status update to Active
        * add VT in description (format: "VTxxxxx | GSxxxx - TEST/PROD")
        * add tenant
        * get parent prefix + role
        * add Tag
    - ip address:
        * status update to Active
        * add VT in description (format: "VOIPT - VTxxxx | GSxxxxx - PROD VRRP SBC")
        * add tenant
        * add Tag
    """
    tenant = Tenant.objects.filter(name=tenant_name).first()

    return_values = dict(
        vlan=str(vlan.vid),
        serviceid=serviceid.name,
        tenant=dict(accountid=tenant.name, account_name=tenant.description.split(" - ")[-1]),
        prefixes=[
            # dict(
            #     environment=None,
            #     prefix=None,
            # )
        ],
        ip_addresses=[
            # dict(
            #     address=None,
            #     description=None,
            # )
        ],
    )

    active_status_obj = Status.objects.filter(name="Active").first()

    # find all the related prefixes and ip addresses and update everything
    prefixes = vlan.prefixes.all()
    ip_addresses = []

    # go over all linked prefixes
    for prefix in prefixes:
        parent_prefix_role = prefix.parent.role.name
        environment = "TEST" if "TEST" in parent_prefix_role else "PROD"
        return_values["prefixes"].append(dict(environment=environment, prefix=str(prefix.prefix)))

        for ip in prefix.ip_addresses.all():
            return_values["ip_addresses"].append(dict(ip_address=str(ip.address), description=ip.description))
            ip_addresses.append(ip)
            ip.status = active_status_obj
            ip.tags.add(serviceid)
            ip.tenant = tenant
            ip.description = f"VOIPT - {serviceid.name} | {ip.description.split(' - ')[-1]}"
            ip.save()

        prefix.status = active_status_obj
        prefix.tags.add(serviceid)
        prefix.tenant = tenant
        prefix.description = f"VOIPT - {serviceid.name} - {environment}"
        prefix.save()

    vlan.status = active_status_obj
    vlan.tags.add(serviceid)
    vlan.tenant = tenant
    vlan.description = f"{serviceid.name} | VOIP Trunking"
    vlan.save()

    return return_values


def security_package_search_first_free_vlan() -> VLAN:
    """Finds the first abailable Security Package vlan.

    We loop over the VOIP_TRUNK_VLAN_RANGES until we can find the first free (reserved) vlan.
    This Vlan is returned.
    """
    all_free_vlans = get_all_reserved_security_package_vlans()
    available_vlans = [int(vlan.vid) for vlan in all_free_vlans]
    vlan_found = None
    for _range in SECURITY_PACKAGE_VLAN_RANGES:
        _from, _to = _range.split(":")
        for _vid in range(int(_from), int(_to) + 1):
            if _vid in available_vlans:
                vlan_found = _vid
                break

    if not vlan_found:
        raise Exception(f"Unable to find a free (reserved) vlan within these ranges: {SECURITY_PACKAGE_VLAN_RANGES}")

    vlan_obj = next(filter(lambda x: x.vid == vlan_found, list(all_free_vlans)))

    return vlan_obj


def reserve_security_package_vlan(
    vlan: VLAN,
    serviceid: Tag,
    # tenant: Tenant,
    tenant_name: str,
    lan_prefix_length: str,
) -> dict:
    """Block a Security Package vlan and matching prefixes and ip addresses.

    All vlans, prefixes and ip-addresses are pre-defined in Nautobot for Security Package.
    We just need to allocate the Tag to each resource in order to link
    it with the service.

    Actions to do:

    - vlan:
        * status update to Active
        * description update (format: "VTxxxx - Security Package")
        * add tenant
        * add Tag

    - prefix:
        * status update to Active
        * add VT in description (format: "VTxxxxx | GSxxxx - Security Package")
        * add tenant
        * get parent prefix + role
        * add Tag
    - ip address:
        * status update to Active
        * add VT in description (format: "VTxxxx | GSxxxxx - Security Package")
        * add tenant
        * add Tag
    """
    # print("1" * 30)

    tenant = Tenant.objects.filter(name=tenant_name).first()

    return_values = dict(
        vlan=str(vlan.vid),
        serviceid=serviceid.name,
        tenant=dict(accountid=tenant.name, account_name=tenant.description.split(" - ")[-1]),
        internet_lan_range="",
        prefixes=[
            # dict(
            #     environment=None,
            #     prefix=None,
            # )
        ],
        ip_addresses=[
            # dict(
            #     address=None,
            #     description=None,
            # )
        ],
    )

    active_status_obj = Status.objects.filter(name="Active").first()
    # print("2" * 30)

    # find a new internet LAN range
    parent_internet_prefix = get_prefix_with_role(SECURITY_PACKAGE_LAN_PREFIX_ROLE)
    new_lan_prefix = get_free_subnet_from_prefix(parent_internet_prefix, int(lan_prefix_length))
    new_subnet = None

    if new_lan_prefix:
        new_subnet = Prefix.objects.create(
            prefix=str(new_lan_prefix),
            type="network",
            status=active_status_obj,
            description=f"{serviceid.name} - Security Package",
            tenant=tenant,
        )
        new_subnet.tags.add(serviceid)
        return_values["internet_lan_range"] = str(new_subnet.prefix)

    # find a new internet LAN range
    parent_internet_prefix = get_prefix_with_role(SECURITY_PACKAGE_LAN_PREFIX_ROLE)
    new_lan_prefix = get_free_subnet_from_prefix(parent_internet_prefix, int(lan_prefix_length))
    new_subnet = None

    if new_lan_prefix:
        new_subnet = Prefix.objects.create(
            prefix=str(new_lan_prefix),
            type="network",
            status=active_status_obj,
            description=f"{serviceid.name} - Security Package",
            tenant=tenant,
        )
        new_subnet.tags.add(serviceid)
        return_values["internet_lan_range"] = str(new_subnet.prefix)

    # find all the related prefixes and ip addresses and update everything
    prefixes = vlan.prefixes.all()
    ip_addresses = []

    # go over all linked prefixes
    for prefix in prefixes:
        # parent_prefix_role = prefix.parent.role.name
        return_values["prefixes"].append(str(prefix.prefix))

        for ip in prefix.ip_addresses.all():
            return_values["ip_addresses"].append(dict(ip_address=str(ip.address), description=ip.description))
            ip_addresses.append(ip)
            ip.status = active_status_obj
            ip.tags.add(serviceid)
            ip.tenant = tenant
            # print("3" * 30)
            ip.description = f"{serviceid.name} - {ip.description} - Security Package"
            # print("4" * 30)
            ip.save()

        prefix.status = active_status_obj
        prefix.tags.add(serviceid)
        prefix.tenant = tenant
        prefix.description = f"{serviceid.name} - Security Package"
        prefix.save()

    vlan.status = active_status_obj
    vlan.tags.add(serviceid)
    vlan.tenant = tenant
    vlan.description = f"{serviceid.name} - Security Package"
    vlan.save()

    return return_values
