"""Functions related to VRFs."""

import re
from nautobot.ipam.models import VRF
from nautobot.ipam.models import RouteTarget
from nautobot.tenancy.models import Tenant
import logging
from django.db.models import Value, IntegerField
from django.db.models.functions import Replace, Cast

logger = logging.getLogger(__name__)

VRF_TYPES = (
    ("any", "Any"),
    ("hub", "Hub"),
    ("spoke", "Spoke"),
)

RD_START_VALUE = 100
RT_START_VALUE = 100


def create_new_vrf_name(proposed_vrf: str) -> str:
    """Returns a new VRF name based on a proposed name.

    If the proposed name already has an index attached, then an exact match
    is searched in the DB, if found, an error is generated.
    If not found, then the proposed name will be returned as new VRF name.

    If there is no index in the proposed name then we search the DB for
    an existing VRF, if found then the existing index will be increased.
    If not found then index will be 1.

    Args:
    ----
        proposed_vrf (str): the proposed VRF name, either with or without index
                            ex. MY_NEW_VRF or MY_NEW_VRF_0001
        model (VRF): the VRF db model

    Returns:
    -------
        vrf name (str)
    """
    vrf: str = proposed_vrf
    vrf_index: int | None = None

    # find the index in the VRF
    if re.match(r"^.*_\d{1,4}$", vrf):
        vrf_index = int(vrf.split("_")[-1])
        vrf = "_".join(vrf.split("_")[:-1]).upper()

    # check the DB if the exact vrf exists
    if vrf_index:
        vrfs_in_db = VRF.objects.filter(name__iexact=vrf).all()
        if vrfs_in_db:
            raise Exception(f"VRF '{vrf}' already exists")

    # check if VRFs with the same name already exist and find
    # the highest VRF index
    if not vrf_index:
        vrf_index = 1
        vrfs_in_db = VRF.objects.filter(name__icontains=vrf).all()
        for _vrf in vrfs_in_db:
            _idx = _vrf.name.split("_")[-1]
            try:
                _idx = int(_idx)
                if _idx >= vrf_index:
                    vrf_index = _idx + 1
            except Exception:
                pass
    new_vrf = f"{vrf}_{str(vrf_index).zfill(4)}"
    logger.debug(f"generated VRF name = {new_vrf}")

    return new_vrf


def get_next_rd() -> int:
    """Searches the VRF DB and finds the next available RD value.

    RD values are always simple integers and the int
    is always increasing.

    The next available RD will be the highest rd value + 1.

    Returns
    -------
        new rd value (int)
    """
    new_rd: int = RD_START_VALUE - 1

    vrfs_in_db = VRF.objects.annotate(rdval=Cast("rd", IntegerField())).order_by("rdval").reverse().first()
    if vrfs_in_db:
        try:
            _rd = int(vrfs_in_db.rd) + 1
            new_rd = _rd
        except Exception:
            pass

    logger.debug(f"generated new rd value: {new_rd}")
    return new_rd


def get_next_route_target() -> str:
    """Searches the VRF DB and finds the next available RT value.

    RT values are always in the format of "47377:<int>" and the <int>
    value is always increasing.

    The next available RT will be the highest RT value + 1.

    Returns
    -------
        new route-target value (str)
    """
    new_rt = RT_START_VALUE - 1
    rts_in_db = (
        RouteTarget.objects.filter(name__contains="47377:")
        .annotate(rtval=Cast(Replace("name", Value("47377:"), Value("")), IntegerField()))
        .order_by("rtval")
        .reverse()
        .first()
    )
    if rts_in_db:
        try:
            _rt = int(rts_in_db.name.split(":")[1])
            new_rt = _rt + 1
        except Exception:
            pass

    logger.debug(f"generated new route-target: 47377:{new_rt}")

    return f"47377:{new_rt}"


def generate_route_targets_for_vrf_type(  # noqa: D417
    vrf_type: str,
    parent_vrf: VRF | None = None,
) -> dict:
    """Generates new import/export route-targets for a certain VRF type.

    A vrf type = any, hub or spoke.

    In case of a hub both "any" and hub route-targets are generated.
    In case of spoke, a hub vrf is requred and the (inverse) hub route-targets are used.

    Args:
    ----
        vrf_type (VRF_TYPES)
        parent_vrf (VRF)

    Returns:
    -------
        dict(
            import=<list of import route-targets>,
            export=<list of export route-targets>,
        )
    """
    new_import_rt: list[str] = []
    new_export_rt: list[str] = []

    if vrf_type == "any" or vrf_type == "hub":
        # find a new available route-target from DB
        new_rt: str = get_next_route_target()
        new_import_rt.append(new_rt)
        new_export_rt.append(new_rt)

        if vrf_type == "hub":
            logger.debug(f"new_rt value:{new_rt}")
            new_rt_int = int(new_rt.split(":")[1])
            new_import_rt.append(f"47377:{new_rt_int+1}")
            new_export_rt.append(f"47377:{new_rt_int+2}")

    if vrf_type == "spoke":
        if not parent_vrf:
            raise Exception("You have to choose a valid parent VRF when creating a new Spoke VRF")

        parent_import_rt = [str(rt) for rt in parent_vrf.import_targets.all()]
        parent_export_rt = [str(rt) for rt in parent_vrf.export_targets.all()]
        if not parent_import_rt or not parent_export_rt:
            raise Exception(f"The parent VRF '{parent_vrf.name}' does not have valid HUB route-targets defined")

        new_import_rt = list(set(parent_export_rt) - set(parent_import_rt))
        new_export_rt = list(set(parent_import_rt) - set(parent_export_rt))

    return {
        "import": new_import_rt,
        "export": new_export_rt,
    }


def create_new_vrf(
    name: str,
    vrf_type: str,
    account: Tenant,
    parent_vrf: VRF | None = None,
) -> VRF:
    """Creates a new VRF object in the DB.

    This

    Args:
    ----
        name (str): new VRF name
        vrf_type (str): any, hub, spoke
        account (Tenant): tenant
        parent_vrf (VRF) : parent VRF in case vrf_type is 'spoke'
    """
    # new VRF name
    new_vrf_name = create_new_vrf_name(name)
    new_rd = get_next_rd()
    new_route_targets = generate_route_targets_for_vrf_type(
        vrf_type=vrf_type,
        parent_vrf=parent_vrf,
    )

    # create new route-targets
    import_targets: list[RouteTarget] = []
    export_targets: list[RouteTarget] = []
    if vrf_type == "any" or vrf_type == "hub":
        for rt in set(new_route_targets["import"] + new_route_targets["export"]):
            try:
                _rt = RouteTarget.objects.create(name=rt, tenant=account, description=new_vrf_name)
                if rt in new_route_targets["import"]:
                    import_targets.append(_rt)
                if rt in new_route_targets["export"]:
                    export_targets.append(_rt)
            except Exception as e:
                raise Exception(f"unable to create route-targets: {e}")

    if vrf_type == "spoke":
        for rt in new_route_targets["import"]:
            _rt = RouteTarget.objects.filter(name=rt).first()
            import_targets.append(_rt)
        for rt in new_route_targets["export"]:
            _rt = RouteTarget.objects.filter(name=rt).first()
            export_targets.append(_rt)

    try:
        vrf = VRF.objects.create(
            name=new_vrf_name,
            rd=new_rd,
            tenant=account,
            description=f"type:{vrf_type.upper()}",
            _custom_field_data=dict(cf_vrf_type=vrf_type.upper()),
        )
        for rt in import_targets:
            vrf.import_targets.add(rt)
        for rt in export_targets:
            vrf.export_targets.add(rt)
        return vrf
    except Exception as e:
        raise Exception(f"unable to create new VRF: {str(e)}")
